//+------------------------------------------------------------------+
//|                                                        ct_10.mq5 |
//|                                  Copyright 2022, MetaQuotes Ltd. |
//|                                             https://www.mql5.com |
//+------------------------------------------------------------------+
#property copyright "Copyright 2022, MetaQuotes Ltd."
#property link      "https://www.mql5.com"
#property version   "1.00"
#property script_show_inputs
#include <ct_10.mqh>
#include <Math\Stat\Math.mqh>
int                        __LOOKBACKS[8] = {1,2,3,4,5,6,7,8};
ENUM_TIMEFRAMES            __TIMEFRAMES[8] = {PERIOD_M12,PERIOD_M15,PERIOD_M20,PERIOD_M30,PERIOD_H1,PERIOD_H2,PERIOD_H3,PERIOD_H4};
ENUM_APPLIED_PRICE         __APPLIEDPRICES[4] = { PRICE_MEDIAN, PRICE_TYPICAL, PRICE_OPEN, PRICE_CLOSE };
string                     __INDICATORS[2] = { "RSI", "BOLLINGER_BANDS" };
string                     __DECISIONS[2] = { "TREND", "RANGE" };


MqlRates __r[];
//+------------------------------------------------------------------+
//| INPUTS                                                           |
//+------------------------------------------------------------------+
input int __timeframe_group_size=21;
input int __lookback_group_size=5;
input int __appliedprice_group_size=21;
input int __indicator_group_size=5;
input int __decision_group_size=21;
//+------------------------------------------------------------------+
//| Script program start function                                    |
//+------------------------------------------------------------------+
void OnStart()
  {
//---
      //declaring a monoid
      CMonoid<int> _m;
      
//+------------------------------------------------------------------+
//| Check for inversion                                              |
//+------------------------------------------------------------------+
      CMonoidGroup<int> _vg;        //valid inversion group
      CMonoidGroup<int> _ig;        //invalid inversion group
      
      _vg.Weights(5);             //set group size
      _ig.Weights(5);             //set group size
      for(int i=0;i<5;i++)
      { 
         CElement<int> _ve;_ve.Cardinality(1); _ve.Set(0,i-2);
         _vg.Set(i,_ve,true);      //set element
         _vg.SetWeight(i,double(i-2));  //set weight
         
         CElement<int> _ie;_ie.Cardinality(1); _ie.Set(0,i);
         _ig.Set(i,_ie,true);      //set element
         _ig.SetWeight(i,double(i));   //set weight
      }
      
      _vg.Operation(OP_ADD);      //set monoid operation to add
      _vg.Identity(2);            //set identity element index to 2
      
      _ig.Operation(OP_ADD);      //set monoid operation to add
      _ig.Identity(2);            //set identity element index to 2 as above or any index
      
      printf(" it is: "+string(_vg.HasInversion())+", vg has inversion, given the weights. ");
      ArrayPrint(_vg.weights,0,",",0,WHOLE_ARRAY,ARRAYPRINT_LIMIT);
      
      printf(" it is: "+string(_ig.HasInversion())+", ig has inversion, given the weights. ");
      ArrayPrint(_ig.weights,0,",",0,WHOLE_ARRAY,ARRAYPRINT_LIMIT);
      
//+------------------------------------------------------------------+
//| Set weights                                                      |
//+------------------------------------------------------------------+

      //time frames
      CMonoidGroup<ENUM_TIMEFRAMES> _t; _t.Weights(8);
      for(int i=0;i<8;i++)
      {
         CElement<ENUM_TIMEFRAMES> _e;_e.Cardinality(1); _e.Set(0,__TIMEFRAMES[i]); _t.Set(i,_e);    
      }
      
      WeighTimeframes(_t);
      printf(" with an input size of: "+IntegerToString(__timeframe_group_size)+" timeframe weights, and their respective monoid action values (group normalised) are: ");
      ArrayPrint(_t.weights,0,",",0,WHOLE_ARRAY,ARRAYPRINT_LIMIT);
      
      int _t_group[];ArrayResize(_t_group,__timeframe_group_size);
      for(int i=0;i<__timeframe_group_size;i++){ _t_group[i]=i; }
      CMonoidAction<ENUM_TIMEFRAMES,int> _ta; _ta.Weights(8); _ta.SetModulo(__timeframe_group_size); _ta.set.Weights(8);
      for(int i=0;i<8;i++){ _ta.set.weights[i]=_t.weights[i]; }
      for(int i=0;i<8;i++)
      {
         //apply action
         CElement<int> _e;_e.Cardinality(1); 
         int _value=_ta.OperateModulo(i,__timeframe_group_size);
         _e.Set(0,_value); _ta.set.Set(i,_e); _ta.set.SetWeight(i,_value);   
      }
      printf(PrintDomain(_ta.set));
      printf(" and action group values (relative set) are: ");
      ArrayPrint(_t_group,0,",",0,WHOLE_ARRAY,ARRAYPRINT_LIMIT);
      
      //
      //lookbacks
      CMonoidGroup<int> _l; _l.Weights(8);
      for(int i=0;i<8;i++)
      {
         CElement<int> _e;_e.Cardinality(1); _e.Set(0,__LOOKBACKS[i]); _l.Set(i,_e);    
      }
      
      //if 1-hour timeframe is selected at previous monoid
      ENUM_TIMEFRAMES _tf=PERIOD_H1;
      WeighLookback(_l,_tf);
      printf(" with an input size of: "+IntegerToString(__lookback_group_size)+" lookback weights, and their respective monoid action values (group normalised) are: ");
      ArrayPrint(_l.weights,0,",",0,WHOLE_ARRAY,ARRAYPRINT_LIMIT);
      
      int _l_group[];ArrayResize(_l_group,__lookback_group_size);
      for(int i=0;i<__lookback_group_size;i++){ _l_group[i]=i; }
      CMonoidAction<int,int> _la; _la.Weights(8); _la.SetModulo(__lookback_group_size); _la.set.Weights(8);
      for(int i=0;i<8;i++){ _la.set.weights[i]=_l.weights[i]; }
      for(int i=0;i<8;i++)
      {
         //apply action
         CElement<int> _e;_e.Cardinality(1); 
         int _value=_la.OperateModulo(i,__lookback_group_size);
         _e.Set(0,_value); _la.set.Set(i,_e); _la.set.SetWeight(i,_value);   
      }
      printf(PrintDomain(_la.set));//ArrayPrint(_la.weights);
      printf(" and action group values (relative set) are: ");
      ArrayPrint(_l_group,0,",",0,WHOLE_ARRAY,ARRAYPRINT_LIMIT);
      
      //
      //appliedprices
      CMonoidGroup<ENUM_APPLIED_PRICE> _a; _a.Weights(4); _a.Cardinality(4);
      for(int i=0;i<4;i++)
      {
         CElement<ENUM_APPLIED_PRICE> _e;_e.Cardinality(1); _e.Set(0,__APPLIEDPRICES[i]); _a.Set(i,_e);    
      }
      
      //if 8 lookback period is selected at previous monoid
      int _lb=8;
      WeighAppliedprice(_a,_tf,_lb);
      printf(" with an input size of: "+IntegerToString(__appliedprice_group_size)+" appliedprice weights, and their respective monoid action values (group normalised) are: ");
      ArrayPrint(_a.weights,0,",",0,WHOLE_ARRAY,ARRAYPRINT_LIMIT);
      
      int _a_group[];ArrayResize(_a_group,__appliedprice_group_size);
      for(int i=0;i<__appliedprice_group_size;i++){ _a_group[i]=i; }
      CMonoidAction<ENUM_APPLIED_PRICE,int> _aa; _aa.Weights(4); _aa.SetModulo(__appliedprice_group_size); _aa.set.Weights(4);
      for(int i=0;i<4;i++){ _aa.set.weights[i]=_a.weights[i]; }
      for(int i=0;i<4;i++)
      {
         //apply action
         CElement<int> _e;_e.Cardinality(1); 
         int _value=_aa.OperateModulo(i,__appliedprice_group_size);
         _e.Set(0,_value); _aa.set.Set(i,_e); _aa.set.SetWeight(i,_value);   
      }
      printf(PrintDomain(_aa.set));//ArrayPrint(_aa.weights);
      printf(" and action group values (relative set) are: ");
      ArrayPrint(_a_group,0,",",0,WHOLE_ARRAY,ARRAYPRINT_LIMIT);
   }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void WeighTimeframes(CMonoidGroup<ENUM_TIMEFRAMES> &G)
   {
      for(int i=0;i<G.Cardinality();i++)
      {
         ResetLastError();
         int _value=0;
         ArrayResize(__r,3);//ArrayInitialize(_buffer,0.0);//ArraySetAsSeries(_buffer,true);
         if(CopyRates(_Symbol,__TIMEFRAMES[i],0,3,__r)>=3)
         {
            _value=int(round(10000.0*fabs(__r[0].close-__r[1].close)/fmax(_Point,fabs(__r[0].close-__r[1].close)+fabs(__r[1].close-__r[2].close))));
         }
         else{ printf(__FUNCSIG__+" Failed to copy: "+EnumToString(__TIMEFRAMES[i])+" close prices. err: "+IntegerToString(GetLastError())); }
         
         ResetLastError();
         if(!G.SetWeight(i,_value))
         {
            printf(__FUNCSIG__+" Failed to assign element at index: "+IntegerToString(i)+", for lookback. ERR: "+IntegerToString(GetLastError()));
         }
      }
   }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void WeighLookback(CMonoidGroup<int> &G,ENUM_TIMEFRAMES &Timeframe)
   {
      for(int i=0;i<G.Cardinality();i++)
      {
         ResetLastError();
         int _value=0;
         int _period=(__LOOKBACKS[i]*PeriodSeconds(Timeframe))/PeriodSeconds(_Period);
         
         ArrayResize(__r,_period+_period+1);
         if(CopyRates(_Symbol,Timeframe,0,_period+_period+1,__r)>=_period+_period+1)
         {
            _value=int(round(10000.0*fabs(__r[0].close-__r[_period].close)/(fabs(__r[0].close-__r[_period].close)+fabs(__r[_period].close-__r[_period+_period].close))));
         }
         else{ printf(__FUNCSIG__+" Failed to copy: "+IntegerToString(__LOOKBACKS[i])+" close prices. err: "+IntegerToString(GetLastError())); }
         
         ResetLastError();
         if(!G.SetWeight(i,_value))
         {
            printf(__FUNCSIG__+" Failed to assign element at index: "+IntegerToString(i)+", for lookback. ERR: "+IntegerToString(GetLastError()));
         }
      }
   }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void WeighAppliedprice(CMonoidGroup<ENUM_APPLIED_PRICE> &G,ENUM_TIMEFRAMES Timeframe,int Lookback)
   {
      for(int i=0;i<G.Cardinality();i++)
      {
         ResetLastError();
         double _value=0.0;
         if(__APPLIEDPRICES[i]==PRICE_CLOSE)
         {
            double _buffer[];ArrayResize(_buffer,Lookback);ArrayInitialize(_buffer,0.0);ArraySetAsSeries(_buffer,true);
            if(CopyClose(_Symbol,Timeframe,0,Lookback,_buffer)>=Lookback)
            {
               _value=MathStandardDeviation(_buffer);
            }
            else{ printf(__FUNCSIG__+" Failed to copy: "+EnumToString(Timeframe)+" close prices. err: "+IntegerToString(GetLastError())); }
         }
         else if(__APPLIEDPRICES[i]==PRICE_OPEN)
         {
            double _buffer[];ArrayResize(_buffer,Lookback);ArrayInitialize(_buffer,0.0);ArraySetAsSeries(_buffer,true);
            if(CopyOpen(_Symbol,Timeframe,0,Lookback,_buffer)>=Lookback)
            {
               _value=MathStandardDeviation(_buffer);
            }
            else{ printf(__FUNCSIG__+" Failed to copy: "+EnumToString(Timeframe)+" open prices. err: "+IntegerToString(GetLastError())); }
         }
         else if(__APPLIEDPRICES[i]==PRICE_MEDIAN)
         {
            double _buffer[];ArrayResize(_buffer,Lookback);ArrayInitialize(_buffer,0.0);ArraySetAsSeries(_buffer,true);
            double _h[];ArrayResize(_h,Lookback);ArrayInitialize(_h,0.0);
            double _l[];ArrayResize(_l,Lookback);ArrayInitialize(_l,0.0);
            if(CopyHigh(_Symbol,Timeframe,0,Lookback,_h)>=Lookback && CopyLow(_Symbol,Timeframe,0,Lookback,_l)>=Lookback)
            {
               for(int ii=0;ii<Lookback;ii++)
               {
                  _buffer[ii]=0.5*(_h[ii]+_l[ii]);
               }
               _value=MathStandardDeviation(_buffer);
            }
            else{ printf(__FUNCSIG__+" Failed to copy: "+EnumToString(Timeframe)+" high and/or low prices. err: "+IntegerToString(GetLastError())); }
         }
         else if(__APPLIEDPRICES[i]==PRICE_TYPICAL)
         {
            double _buffer[];ArrayResize(_buffer,Lookback);ArrayInitialize(_buffer,0.0);ArraySetAsSeries(_buffer,true);
            double _h[];ArrayResize(_h,Lookback);ArrayInitialize(_h,0.0);
            double _l[];ArrayResize(_l,Lookback);ArrayInitialize(_l,0.0);
            double _c[];ArrayResize(_c,Lookback);ArrayInitialize(_c,0.0);
            if(CopyHigh(_Symbol,Timeframe,0,Lookback,_h)>=Lookback && CopyLow(_Symbol,Timeframe,0,Lookback,_l)>=Lookback && CopyClose(_Symbol,Timeframe,0,Lookback,_c)>=Lookback)
            {
               for(int ii=0;ii<Lookback;ii++)
               {
                  _buffer[ii]=(1.0/3.0)*(_h[ii]+_l[ii]+_c[ii]);
               }
               _value=MathStandardDeviation(_buffer);
            }
            else{ printf(__FUNCSIG__+" Failed to copy: "+EnumToString(Timeframe)+" high and/or low prices. err: "+IntegerToString(GetLastError())); }
         }
         else if(__APPLIEDPRICES[i]==PRICE_WEIGHTED)
         {
            double _buffer[];ArrayResize(_buffer,Lookback);ArrayInitialize(_buffer,0.0);ArraySetAsSeries(_buffer,true);
            double _h[];ArrayResize(_h,Lookback);ArrayInitialize(_h,0.0);
            double _l[];ArrayResize(_l,Lookback);ArrayInitialize(_l,0.0);
            double _c[];ArrayResize(_c,Lookback);ArrayInitialize(_c,0.0);
            if(CopyHigh(_Symbol,Timeframe,0,Lookback,_h)>=Lookback && CopyLow(_Symbol,Timeframe,0,Lookback,_l)>=Lookback && CopyClose(_Symbol,Timeframe,0,Lookback,_c)>=Lookback)
            {
               for(int ii=0;ii<Lookback;ii++)
               {
                  _buffer[ii]=(1.0/4.0)*(_h[ii]+_l[ii]+_c[ii]+_c[ii]);
               }
               _value=MathStandardDeviation(_buffer);
            }
            else{ printf(__FUNCSIG__+" Failed to copy: "+EnumToString(Timeframe)+" high and/or low prices. err: "+IntegerToString(GetLastError())); }
         }
         
         _value/=_Point;
         
         int _v=int(round(10000*_value));
         
         ResetLastError();
         if(!G.SetWeight(i,_v))
         {
            printf(__FUNCSIG__+" Failed to assign element at index: "+IntegerToString(i)+", for lookback. ERR: "+IntegerToString(GetLastError()));
         }
      }
   }
//+------------------------------------------------------------------+
